"""Provides common base classes that allow different modules to talk to each other."""
from dataclasses import dataclass
from itertools import chain
from typing import Sequence

from pydantic.main import BaseModel
from pydicom.dataset import Dataset

from dicomtrolley.exceptions import DICOMTrolleyError


class DICOMObjectLevels:
    """DICOM uses a study->series->instance structure"""

    STUDY = "Study"
    SERIES = "Series"
    INSTANCE = "Instance"


class DICOMObjectReference:
    def __init__(self, study_uid, series_uid=None, instance_uid=None):
        """Designates a study, series or instance uid by uids

        Made this to remove all

        Parameters
        ----------
        study_uid: str
        series_uid: str
        instance_uid: str

        Raises
        ------
        ValueError
            If data for the given parameters does not exist in tree
        """
        if instance_uid and not series_uid:
            raise ValueError(
                f"Instance was given ({instance_uid}) but series was not. I can "
                f"not insert this into a study/series/instance tree"
            )

        self.study_uid = study_uid
        self.series_uid = series_uid
        self.instance_uid = instance_uid

    @property
    def level(self):
        """Does this reference point to Study, Series or Instance?"""
        if self.instance_uid:
            return DICOMObjectLevels.INSTANCE
        elif self.series_uid:
            return DICOMObjectLevels.SERIES
        else:
            return DICOMObjectLevels.STUDY

    def __str__(self):
        return (
            f"{self.level} reference {self.study_uid} -> {self.series_uid} "
            f"-> {self.instance_uid}"
        )


class DICOMObject(BaseModel):
    """An object in the DICOM world. Base for Study, Series, Instance.

    dicomtrolley search methods always return instances based on DICOMObject
    dicomtrolley download methods take instances based on DICOMObject as input
    """

    class Config:
        arbitrary_types_allowed = True  # allows the use of Dataset type below

    uid: str
    data: Dataset

    def __str__(self):
        return type(self).__name__ + " " + self.uid

    def children(self):
        """

        Returns
        -------
        Sequence[DICOMObject]
            All direct child objects or empty list if none. Useful for iterating
        """
        raise NotImplementedError()

    def root(self):
        """The parent or parents parent or etc.

        Returns
        -------
        DICOMObject
        """
        raise NotImplementedError()

    def all_instances(self):
        """

        Returns
        -------
        List[Instance]
            All instances contained in this object
        """
        raise NotImplementedError()

    def reference(self) -> DICOMObjectReference:
        """Return a Reference to this object using uids"""
        raise NotImplementedError()


class Instance(DICOMObject):
    parent: "Series"

    def all_instances(self):
        """A list containing this instance itself. To match other signatures"""
        return [self]

    def children(self):
        return []

    def root(self):
        return self.parent.parent

    def reference(self) -> DICOMObjectReference:
        """Return a Reference to this object using uids"""
        return DICOMObjectReference(
            study_uid=self.parent.parent.uid,
            series_uid=self.parent.uid,
            instance_uid=self.uid,
        )


class Series(DICOMObject):
    instances: Sequence[Instance]
    parent: "Study"

    def __getitem__(self, instance_uid):
        return self.get(instance_uid)

    def all_instances(self):
        """Each instance contained in this series"""
        return self.instances

    def children(self):
        return self.instances

    def root(self):
        return self.parent

    def get(self, instance_uid: str) -> Instance:
        """Get instance with this uid"""
        try:
            return next(x for x in self.instances if x.uid == instance_uid)
        except StopIteration as e:
            raise KeyError(
                f'instance with uid "{instance_uid}" not found in series'
            ) from e

    def reference(self) -> DICOMObjectReference:
        """Return a Reference to this object using uids"""
        return DICOMObjectReference(
            study_uid=self.parent.uid, series_uid=self.uid
        )


class Study(DICOMObject):
    series: Sequence[Series]

    def __getitem__(self, series_uid) -> Series:
        return self.get(series_uid)

    def all_instances(self):
        """Return each instance contained in this study"""
        return list(chain(*(x.instances for x in self.series)))

    def children(self):
        return self.series

    def root(self):
        return self

    def get(self, series_uid: str) -> Series:
        """Get series with this uid"""
        try:
            return next(x for x in self.series if x.uid == series_uid)
        except StopIteration as e:
            raise KeyError(
                f'series with uid "{series_uid}" not found in study'
            ) from e

    def reference(self) -> DICOMObjectReference:
        """Return a Reference to this object using uids"""
        return DICOMObjectReference(study_uid=self.uid)


@dataclass
class InstanceReference:
    """All information needed to download a single slice (SOPInstance)"""

    study_instance_uid: str
    series_instance_uid: str
    sop_instance_uid: str

    def __str__(self):
        return f"InstanceReference {self.sop_instance_uid}"


class Downloader:
    """Something that can fetch DICOM instances. Base class"""

    def get_dataset(self, instance: InstanceReference):
        """Get DICOM dataset for the given instance (slice)

        Raises
        ------
        DICOMTrolleyError
            If getting does not work for some reason

        Returns
        -------
        Dataset
            A pydicom dataset
        """
        raise NotImplementedError

    def datasets(self, instances: Sequence[InstanceReference]):
        """Retrieve each instance

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        for instance in instances:
            yield self.get_dataset(instance)

    def datasets_async(
        self, instances: Sequence[InstanceReference], max_workers=None
    ):
        """Retrieve each instance in multiple threads

        Parameters
        ----------
        instances: Sequence[InstanceReference]
            Retrieve dataset for each of these instances
        max_workers: int, optional
            Use this number of workers in ThreadPoolExecutor. Defaults to
            default for ThreadPoolExecutor

        Raises
        ------
        DICOMTrolleyError
            When a server response cannot be parsed as DICOM

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        raise NotImplementedError


class Searcher:
    """Something that can search for DICOM studies. Base class."""

    def find_studies(self, query) -> Sequence[Study]:
        raise NotImplementedError()

    def find_study(self, query) -> Study:
        """Like find_studies, but guarantees exactly one result. Exception if not.

        This method is meant for searches that contain unique identifiers like
        StudyInstanceUID, AccessionNumber, etc.

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        results = self.find_studies(query)
        if len(results) == 0 or len(results) > 1:
            raise DICOMTrolleyError(
                f"Expected exactly one study for query '{query}', but"
                f" found {len(results)}"
            )
        return results[0]

    def find_full_study_by_id(self, study_uid: str) -> Study:
        """Find a single study at image level

        Useful for automatically finding all instances for a study. Meant to be
        implemented in child classes

        Parameters
        ----------
        study_uid: str
            Study to search for

        Returns
        -------
        Study

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        raise NotImplementedError()


Instance.update_forward_refs()  # enables pydantic validation
Series.update_forward_refs()



"""Combines WADO, RAD69, MINT and DICOM-QR to make make getting DICOM studies easy.

Notes
-----
Design choices:

WADO, RAD69, MINT and DICOM-QR modules should be stand-alone. They are not allowed to
use each other's classes. Core has knowledge of all and converts between them if
needed

"""
from pathlib import Path
from typing import List, Sequence, Tuple, Union

from dicomtrolley.core import (
    DICOMObject,
    Downloader,
    Instance,
    InstanceReference,
    Searcher,
    Study,
)
from dicomtrolley.parsing import DICOMObjectTree
from dicomtrolley.types import DICOMDownloadable


class Trolley:
    """Combines a search and download method to get DICOM studies easily"""

    def __init__(
        self, downloader: Downloader, searcher: Searcher, query_missing=True
    ):
        """

        Parameters
        ----------
        downloader: Downloader
            The module to use for downloads
        searcher: Searcher
            The module to use for queries
        query_missing: bool, optional
            if True, Trolley.download() will query for missing DICOM instances. For
            example when passing a Study obtained from a study-level query, which does
            not contain any information on instances
            if False, missing instances will not be downloaded

        """
        self.downloader = downloader
        self.searcher = searcher
        self._searcher_cache = DICOMObjectTree([])
        self.query_missing = query_missing

    def find_studies(self, query) -> List[Study]:
        """Find study information

        Parameters
        ----------
        query:
            Search for these criteria

        Returns
        -------
        List[Study]
        """
        return list(self.searcher.find_studies(query))

    def find_study(self, query) -> Study:
        """Like find study, but returns exactly one result or raises exception.

        For queries using study identifiers such as StudyInstanceUID,AccessionNumber

        Parameters
        ----------
        query
            Search for these criteria. Query documentation for more info.

        Returns
        -------
        Study
        """
        return self.searcher.find_study(query)

    def download(
        self,
        objects: Union[DICOMDownloadable, Sequence[DICOMDownloadable]],
        output_dir,
        use_async=False,
        max_workers=None,
    ):
        """Download the given objects to output dir."""
        if not isinstance(objects, Sequence):
            objects = [objects]  # if just a single item to download is passed
        storage = DICOMStorageDir(output_dir)
        if use_async:
            datasets = self.fetch_all_datasets_async(
                objects=objects, max_workers=max_workers
            )
        else:
            datasets = self.fetch_all_datasets(objects=objects)

        for dataset in datasets:
            storage.save(dataset)

    def fetch_all_datasets(self, objects: Sequence[DICOMDownloadable]):
        """Get full DICOM dataset for all instances contained in objects.

        Returns
        -------
        Iterator[Dataset]
            The downloaded dataset and the context that was used to download it
        """

        yield from self.downloader.datasets(
            self.objects_to_references(objects)
        )

    def objects_to_references(
        self, objects: Sequence[DICOMDownloadable]
    ) -> Sequence[InstanceReference]:
        """Find all instances contained in the given objects. Query for missing"""
        references, dicom_objects = self.split_references(objects)
        if self.query_missing:
            dicom_objects = self.ensure_instances(dicom_objects)
        return references + self.extract_references(dicom_objects)

    @staticmethod
    def split_references(
        objects,
    ) -> Tuple[List[InstanceReference], List[DICOMObject]]:
        references, dicom_objects = [], []
        for item in objects:
            references.append(item) if isinstance(
                item, InstanceReference
            ) else dicom_objects.append(item)
        return references, dicom_objects

    @staticmethod
    def extract_references(
        objects: Sequence[DICOMDownloadable],
    ) -> List[InstanceReference]:
        """Get all individual instances from input.

        A pre-processing step for getting datasets.

        Parameters
        ----------
        objects: list[DICOMDownloadable]
            Any combination of Study, Series and Instance objects

        Returns
        -------
        List[InstanceReference]
            A reference to each instance (slice)
        """
        instances: List[InstanceReference]
        instances = []
        for item in objects:
            if isinstance(item, InstanceReference):
                instances.append(item)
            else:
                instances = instances + [
                    to_instance_reference(x) for x in item.all_instances()
                ]
        return instances

    def ensure_instances(self, objects: Sequence[DICOMObject]):
        """Ensure that all objects contain instances. Perform additional image-level
        queries with searcher if needed.

        Note
        ----
        This method fires additional search queries and might take a long time to
        return depending on the number of objects and missing instances
        """

        def has_instances(item_in):
            return bool(item_in.all_instances())

        # all information on studies, series and instances we have
        self._searcher_cache = DICOMObjectTree(objects)
        cache = self._searcher_cache

        ensured = []
        for item in objects:
            if has_instances(item):
                ensured.append(item)  # No work needed
            else:
                retrieved = cache.retrieve(
                    item.reference()
                )  # have we cached before?
                if has_instances(retrieved):  # yes we have. Add that
                    ensured.append(retrieved)
                else:  # No instances, we'll have to query for them
                    study = self.searcher.find_full_study_by_id(
                        item.root().uid
                    )
                    cache.add_study(study)
                    ensured.append(cache.retrieve(item.reference()))

        self._searcher_cache = DICOMObjectTree([])
        return ensured

    def fetch_all_datasets_async(self, objects, max_workers=None):
        """Get DICOM dataset for each instance given objects using multiple threads.

        Parameters
        ----------
        objects: Sequence[DICOMDownloadable]
            get dataset for each instance contained in these objects
        max_workers: int, optional
            Max number of ThreadPoolExecutor workers to use. Defaults to
            ThreadPoolExecutor default

        Raises
        ------
        DICOMTrolleyError
            If getting or parsing of any instance fails

        Returns
        -------
        Iterator[Dataset, None, None]
            The downloaded dataset and the context that was used to download it
        """

        yield from self.downloader.datasets_async(
            instances=self.objects_to_references(objects),
            max_workers=max_workers,
        )


def to_instance_reference(
    item: Union[Instance, InstanceReference]
) -> InstanceReference:
    """Simplify a more extensive instance to an InstanceReference.

    needed for calls to WADO functions.

    Opted to check for type here as I don't want put this functionality in core.
    """
    if isinstance(item, InstanceReference):
        return item
    else:
        return InstanceReference(
            study_instance_uid=item.parent.parent.uid,
            series_instance_uid=item.parent.uid,
            sop_instance_uid=item.uid,
        )


class DICOMStorageDir:
    """A directory that you can write datasets to."""

    def __init__(self, path: str):
        self.path = path

    def __str__(self):
        return f"DICOMStorageDir at {self.path}"

    def save(self, dataset):
        """Write dataset. Creates sub-folders if needed."""

        slice_path = Path(self.path) / self.generate_path(dataset)
        slice_path.parent.mkdir(parents=True, exist_ok=True)
        dataset.save_as(slice_path)

    def generate_path(self, dataset):
        """A path studyid/seriesid/instanceid to save a slice to."""

        stu_uid = self.get_value(dataset, "StudyInstanceUID")
        ser_uid = self.get_value(dataset, "SeriesInstanceUID")
        sop_uid = self.get_value(dataset, "SOPInstanceUID")
        return Path(self.path) / stu_uid / ser_uid / sop_uid

    @staticmethod
    def get_value(dataset, tag_name):
        """Extract value for use in path. If not found return default."""
        default = "unknown"
        return str(dataset.get(tag_name, default)).replace(".", "_")



"""Implements DICOM QR (Query Retrieve), a method for getting information from a VNA

See http://dicom.nema.org/dicom/2013/output/chtml/part04/sect_C.3.html
"""
from datetime import datetime
from typing import List, Optional

from pydantic.class_validators import validator
from pydantic.main import BaseModel
from pydicom.datadict import tag_for_keyword
from pydicom.dataset import Dataset
from pynetdicom import AE, debug_logger
from pynetdicom.sop_class import PatientRootQueryRetrieveInformationModelFind

from dicomtrolley.core import Searcher, Study
from dicomtrolley.exceptions import DICOMTrolleyError
from dicomtrolley.parsing import DICOMParseTree


class QueryRetrieveLevels:
    """Valid values for DICOMQR. Differs slightly from the MINT version."""

    STUDY = "STUDY"
    SERIES = "SERIES"
    IMAGE = "IMAGE"

    ALL = {STUDY, SERIES, IMAGE}


class DICOMQuery(BaseModel):
    """Things you can search for with DICOM QR.

    Notes
    -----
    * Incomplete: this class implements only a minimal core of DICOM QR search.
      It can be extended should the need arise
    * All string arguments support (*) as a wildcard
    * non-pep8 parameter naming format follows DICOM parameter convention.
      Other parameters match MINTQuery conventions, query parameters are similar.

    """

    # DICOM parameters (used in CFIND directly)
    AccessionNumber: Optional[str] = ""
    Modality: Optional[str] = ""
    PatientID: Optional[str] = ""
    PatientName: Optional[str] = ""
    ProtocolName: Optional[str] = ""
    SeriesDescription: Optional[str] = ""
    StudyDescription: Optional[str] = ""
    StudyID: Optional[str] = ""
    StudyInstanceUID: Optional[str] = ""

    # meta parameters: how to return results
    QueryRetrieveLevel: str = (
        QueryRetrieveLevels.STUDY
    )  # depth to return results at

    # NON DICOM parameters (are transformed into DICOM parameters
    minStudyDate: Optional[datetime]
    maxStudyDate: Optional[datetime]
    includeFields: Optional[List[str]] = []  # which dicom fields to return

    class Config:
        extra = "forbid"  # raise ValueError when passing an unknown keyword to init

    @validator("QueryRetrieveLevel")
    def query_retrieve_value_should_be_valid(
        cls, value, values  # noqa: B902, N805
    ):
        if value not in QueryRetrieveLevels.ALL:
            raise ValueError(
                f'Unknown QueryRetrieveLevel "{value}". '
                f"Allowed:{QueryRetrieveLevels.ALL}"
            )
        return value

    @validator("includeFields", always=True)
    def include_fields_check(cls, include_fields, values):  # noqa: B902, N805
        """Include fields should be valid dicom fields"""
        for field in include_fields:
            if not tag_for_keyword(field):
                raise ValueError(f"{field} is not a valid DICOM keyword")

        return include_fields

    @staticmethod
    def get_default_include_fields(query_retrieve_level):
        """Include fields you definitely want to get back"""
        if query_retrieve_level == QueryRetrieveLevels.STUDY:
            return ["StudyInstanceUID"]
        elif query_retrieve_level == QueryRetrieveLevels.SERIES:
            return ["StudyInstanceUID", "SeriesInstanceUID"]
        elif query_retrieve_level == QueryRetrieveLevels.IMAGE:
            return ["StudyInstanceUID", "SeriesInstanceUID", "SOPInstanceUID"]

    @staticmethod
    def parse_date(date):
        if date:
            return date.strftime("%Y%m%d")
        else:
            return ""

    @classmethod
    def get_study_date(cls, min_study_date, max_study_date):
        """Get value for CFIND parameter StudyDate"""
        min_sd = cls.parse_date(min_study_date)
        max_sd = cls.parse_date(max_study_date)
        if min_sd or max_sd:
            return min_sd + "-" + max_sd
        else:
            return None

    def as_dataset(self):
        """A dataset that can be used as a CFIND query."""

        # remove non-DICOM parameters and replace with DICOM tags based on them
        parameters = {
            x: y for x, y in self.dict().items()
        }  # all params for query
        parameters["StudyDate"] = self.get_study_date(
            parameters.pop("minStudyDate"), parameters.pop("maxStudyDate")
        )

        # add useful default include fields
        default_fields = self.get_default_include_fields(
            parameters["QueryRetrieveLevel"]
        )
        parameters["includeFields"] = list(
            set(parameters["includeFields"]) | set(default_fields)
        )

        ds = Dataset()
        # in CFIND, empty elements are interpreted as 'need to be returned filled'
        for field in parameters.pop("includeFields"):  # for each include field
            if tag_for_keyword(field):
                setattr(ds, field, "")  # add an empty DICOM element
        parameters = {
            x: y for x, y in parameters.items() if y
        }  # Skip None values

        for parameter, value in parameters.items():
            setattr(ds, parameter, value)

        return ds


class DICOMQR(Searcher):
    """A connection to a DICOM QR enabled server."""

    def __init__(
        self, host, port, aet="DICOMTROLLEY", aec="ANY-SCP", debug=False
    ):
        """

        Parameters
        ----------
        host: str
            Hostname of DICOM-QR-enabled server
        port: int
            Port for DICOM-QR
        aet: str, optional
            Application Entity Title - Name of the calling entity (this class).
            Defaults to 'DICOMTROLLEY'
        aec: str, optional
            Application Entity Called - The name of the server you are calling.
            Defaults to 'ANY-SCP'
        debug: bool, optional
            If True, prints debug logging to console. This can be very useful as
            exceptions often do not contain detailed information.
        """
        self.host = host
        self.port = port
        self.aet = aet
        self.aec = aec
        self.debug = debug

    def find_studies(self, query: DICOMQuery):
        """

        Parameters
        ----------
        query: DICOMQuery
            Find arguments matching this query

        Raises
        ------
        DICOMTrolleyError
            When finding fails

        Returns
        -------
        List[Study]
        """
        return self.parse_c_find_response(self.send_c_find(query))

    @staticmethod
    def parse_c_find_response(responses) -> List[Study]:
        """Parse flat list of datasets from CFIND into a study/series/instance tree

        CFIND returns a flat list of datasets on the queries' QueryRetrieveLevel.
        For instance at IMAGE level, there is one dataset for each matching instance.
        Each dataset should contain Series and Study information. Parse this into a
        dicomtrolley study/series/instance tree that can be used as input for
        download functions

        Parameters
        ----------
        responses: Sequence[Dataset]
            Datasets coming from a pydicom cfind query

        Returns
        -------
        List[Study]
            Each study populated with series and instance objects, if provided
        """

        tree = DICOMParseTree()
        for response in responses:
            tree.insert_dataset(response)
        return tree.as_studies()

    def send_c_find(self, query):
        """Perform a CFIND with the given query

        Raises
        ------
        DICOMTrolleyError
            When finding fails
        """
        if self.debug:
            debug_logger()
        ae = AE(ae_title=bytes(self.aet, encoding="utf-8"))
        ae.add_requested_context(PatientRootQueryRetrieveInformationModelFind)

        assoc = ae.associate(
            self.host, self.port, ae_title=bytes(self.aec, encoding="utf-8")
        )
        responses = []
        if assoc.is_established:
            # Send the C-FIND request
            c_find_response = assoc.send_c_find(
                query.as_dataset(),
                PatientRootQueryRetrieveInformationModelFind,
            )
            for (status, identifier) in c_find_response:
                if status:
                    # I don't understand this status.. For now just collect non-None
                    if identifier:
                        responses.append(identifier)

                else:
                    raise DICOMTrolleyError(
                        "Connection timed out, was aborted or"
                        " received invalid response"
                    )

            assoc.release()
        else:
            raise DICOMTrolleyError(
                "Association rejected, aborted or never connected"
            )

        return responses

    def find_full_study_by_id(self, study_uid: str) -> Study:
        return self.find_study(
            DICOMQuery(
                StudyInstanceUID=study_uid,
                QueryRetrieveLevel=QueryRetrieveLevels.IMAGE,
            )
        )



"""Custom types for type hints"""
from typing import Union

from dicomtrolley.core import DICOMObject, InstanceReference

# Anything that can be downloaded by trolley
DICOMDownloadable = Union[DICOMObject, InstanceReference]



"""Retrieve data from servers using the rad69 protocol

see https://gazelle.ihe.net/content/rad-69-retrieve-imaging-document-set
And the corresponding transaction: https://profiles.ihe.net/ITI/TF/Volume2/ITI-43.html
"""

import email.parser
import math
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Sequence, Tuple
from xml.etree import ElementTree

from jinja2.environment import Template
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomBytesIO
from pydicom.filereader import dcmread
from requests.structures import CaseInsensitiveDict
from requests_futures.sessions import FuturesSession

from dicomtrolley.core import Downloader, InstanceReference
from dicomtrolley.exceptions import DICOMTrolleyError
from dicomtrolley.parsing import DICOMParseTree
from dicomtrolley.xml_templates import (
    RAD69_SOAP_REQUEST_TEMPLATE,
    RAD69_SOAP_RESPONSE_ERROR_XPATH,
)


class Rad69(Downloader):
    """A connection to a Rad69 server"""

    def __init__(self, session, url, http_chunk_size=65536):
        """
        Parameters
        ----------
        session: requests.session
            A logged in session over which rad69 calls can be made
        url: str
            rad69 endpoint, including protocol and port. Like https://server:2525/rids
        http_chunk_size: int, optional
            Number of bytes to read each time when streaming chunked rad69 responses.
            Defaults to 64 Kb (65536 bytes)
        """

        self.session = session
        self.url = url
        self.http_chunk_size = http_chunk_size
        self.template = RAD69_SOAP_REQUEST_TEMPLATE
        self.post_headers = {"Content-Type": "application/soap+xml"}

    def datasets(self, instances: Sequence[InstanceReference]):
        """Retrieve all instances by querying rad69 for all in one request

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        response = self.session.post(
            url=self.url,
            headers=self.post_headers,
            data=self.create_instances_request(instances),
            stream=True,
        )
        return self.parse_rad69_response(response)

    def create_instance_request(self, instance: InstanceReference):
        """Create the SOAP xml structure needed to request an instance from a rad69
        server
        """
        return self.create_instances_request(instances=[instance])

    def create_instances_request(self, instances: Sequence[InstanceReference]):
        """Create the SOAP xml structure to request all given instances from server"""
        # Turn instanceReference list back into study level
        tree = DICOMParseTree()
        for instance in instances:
            tree.insert(
                data=[],
                study_uid=instance.study_instance_uid,
                series_uid=instance.series_instance_uid,
                instance_uid=instance.sop_instance_uid,
            )
        studies = tree.as_studies()

        return Template(self.template).render(
            uuid=str(uuid.uuid4()),
            studies=studies,
            transfer_syntax_list=["1.2.840.10008.1.2", "1.2.840.10008.1.2.1"],
        )

    def get_dataset(self, instance: InstanceReference):
        """Get DICOM dataset for the given instance (slice)

        Raises
        ------
        DICOMTrolleyError
            If getting does not work for some reason

        Note
        ----
        Getting single datasets is not efficient in rad69. If you want to loop
        this function, use Rad69.datasets() instead

        Returns
        -------
        Dataset
            A pydicom dataset
        """

        response = self.session.post(
            url=self.url,
            headers=self.post_headers,
            data=self.create_instance_request(instance),
        )
        return list(self.parse_rad69_response(response))[0]

    @staticmethod
    def verify_rad69_response(response):
        """Raise exceptions if this response is not a multi-part rad69 soap response.

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected
        """
        if response.status_code != 200:
            raise DICOMTrolleyError(
                f"Calling {response.url} failed ({response.status_code} - "
                f"{response.reason})\n"
                f"response content was {str(response.content[:300])}"
            )

        # check multipart
        if "Content-type" not in response.headers:
            raise DICOMTrolleyError(
                f"Expected multipart response, but got no content type for this"
                f" response. Start of response: {str(response.content[:300])}"
            )

        if "multipart" not in response.headers["Content-Type"].lower():
            # Maybe server is sending back a valid soap error document.
            Rad69.parse_rad69_soap_error_response(response)

    @staticmethod
    def parse_rad69_soap_error_response(response):
        """Interpret response as rad69 error, raise more helpful exceptions

        Raises
        ------
        DICOMTrolleyError
            if this is a valid rad69 error response. Raise nothing otherwise
        """

        tree = ElementTree.fromstring(response.text)
        errors = tree.findall(RAD69_SOAP_RESPONSE_ERROR_XPATH)
        if not errors:
            raise DICOMTrolleyError(
                f"Could not find any rad69 soap errors in "
                f"response: {response.content[:900]}"
            )
        else:
            raise DICOMTrolleyError(
                f"Rad69 server returns {len(errors)} errors:"
                f" {[str(error.attrib) for error in errors]}"
            )

    def parse_rad69_response(self, response):
        """Create a Dataset out of http response from a rad69 server

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected or if parsing fails

        Returns
        -------
        Iterator[Dataset, None, None]
            All datasets included in this response
        """
        self.verify_rad69_response(response)
        part_stream = HTTPMultiPartStream(
            response, stream_chunk_size=self.http_chunk_size
        )
        soap_part = None
        for part in part_stream:
            if not soap_part:
                soap_part = part  # skip soap part of the response
                continue
            dicom_bytes = part.content
            raw = DicomBytesIO(dicom_bytes)
            try:
                yield dcmread(raw)
            except InvalidDicomError as e:
                raise DICOMTrolleyError(
                    f"Error parsing response as dicom: {e}."
                    f" Response content (first 300 elements) was"
                    f" {str(response.content[:300])}"
                ) from e

    @staticmethod
    def split_instances(instances: Sequence[InstanceReference], num_bins):
        """Split the given instance references into even piles"""
        bin_size = math.ceil(len(instances) / num_bins)
        for i in range(0, len(instances), bin_size):
            yield instances[i : i + bin_size]

    def datasets_async(
        self, instances: Sequence[InstanceReference], max_workers=4
    ):
        """Split instances into chunks and retrieve each chunk in separate thread

        Parameters
        ----------
        instances: Sequence[InstanceReference]
            Retrieve dataset for each of these instances
        max_workers: int, optional
            Use this number of workers in ThreadPoolExecutor. Defaults to
            default for ThreadPoolExecutor

        Notes
        -----
        rad69 allows any number of slices to be combined in one request. The response
        is a chunked multi-part http response with all image data. Requesting each
        slice individually is inefficient. Requesting all slices in one thread might
        limit speed. Somewhere in the middle seems the best bet for optimal speed.
        This function splits all instances between the available workers and lets
        workers process the response streams.

        Raises
        ------
        DICOMTrolleyError
            When a server response cannot be parsed as DICOM

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        # max_workers=None means let the executor figure it out. But for rad69 we
        # still need to determine how many instances to retrieve at once with each
        # worker. Unlimited workers make no sense here. Just use a single thread.
        if max_workers is None:
            max_workers = 1

        with FuturesSession(
            session=self.session,
            executor=ThreadPoolExecutor(max_workers=max_workers),
        ) as futures_session:
            futures = []
            for instance_bin in self.split_instances(instances, max_workers):
                futures.append(
                    futures_session.post(
                        url=self.url,
                        headers=self.post_headers,
                        data=self.create_instances_request(instance_bin),
                    )
                )

            for future in as_completed(futures):
                yield from self.parse_rad69_response(future.result())


def _split_on_find(content, bound):
    """Split content string on a substring"""
    point = content.find(bound)
    return content[:point], content[point + len(bound) :]


def parse_headers(content, encoding):
    string = content.decode(encoding)
    return email.parser.HeaderParser().parsestr(string).items()


class HTMLPart:
    """One part of a multi-part http response, without the boundaries"""

    def __init__(self, content, encoding):
        if not encoding:
            encoding = "utf-8"
        self.encoding = encoding
        headers = {}
        # Split into header section (if any) and the content
        if b"\r\n\r\n" in content:
            first, self.content = _split_on_find(content, b"\r\n\r\n")
            if first != b"":
                headers = parse_headers(first.lstrip(), encoding)
        else:
            raise MultipartContentError("content does not contain CR-LF-CR-LF")
        self.headers = CaseInsensitiveDict(headers)

    @property
    def text(self):
        return self.content.decode(self.encoding)


class HTTPMultiPartStream:
    """Converts a streamed http multipart response into separate parts.

    Main use is as an iterator:

        parts = [x for x in HTTPMultiPartStream(response)]

    This iterator is stateful and can only be called once as it consumes the
    response stream
    """

    def __init__(self, response, stream_chunk_size=65536):
        self.response = response
        self.boundary = self._find_boundary(response)

        self._bytes_iterator = iter(
            response.iter_content(chunk_size=stream_chunk_size)
        )
        self._buffer = b""

    @staticmethod
    def _split_on_find(content, bound):
        point = content.find(bound)
        return content[:point], content[point + len(bound) :]

    @classmethod
    def _find_boundary(cls, multipart_response):
        content_type_info = tuple(
            x.strip()
            for x in multipart_response.headers.get("content-type").split(";")
        )
        mimetype = content_type_info[0]
        if mimetype.split("/")[0].lower() != "multipart":
            raise MultipartContentError(
                f"Unexpected mimetype in content-type: '{mimetype}'"
            )
        for item in content_type_info[1:]:
            attr, value = cls._split_on_find(item, "=")
            if attr.lower() == "boundary":
                return value.strip('"').encode("utf-8")

    def __iter__(self):
        return self

    def __next__(self):
        # is there a part between two boundaries in current buffer?
        part = self.get_next_part_from_buffer()
        if not part:
            while not part:
                self._buffer = self._buffer + next(self._bytes_iterator)
                part = self.get_next_part_from_buffer()
        return HTMLPart(part, encoding=self.response.encoding)

    def get_next_part_from_buffer(self):
        """Return first part in buffer and remove this from buffer"""
        part, rest = self.split_off_first_part(
            self._buffer, b"--" + self.boundary
        )
        if part:
            self._buffer = rest
            return part
        else:
            return None

    @staticmethod
    def split_off_first_part(
        bytes_in: bytes, boundary: bytes
    ) -> Tuple[bytes, bytes]:
        """Find the content between two boundaries.
        Expects bytes_in to start with a boundary

        Returns
        -------
        Tuple[bytes, bytes]
            The content found between first two boundaries and the rest. Rest will
            start with a boundary, content found will not, discarding the boundary
            there.
            If no second boundary is found or bytes_in is empty,
            will return Tuple[b'', bytes_in]

        Raises
        ------
        MultipartContentError
            If bytes_in does not start with a boundary
        """
        if not bytes_in:
            return b"", bytes_in  # Avoid exception for valid empty input
        elif len(bytes_in) < len(boundary):
            return b"", bytes_in  # Not enough bytes to find boundary yet
        elif bytes_in.find(boundary) != 0:
            raise MultipartContentError(
                f"Expected http multipart bytestream to start with "
                f"boundary {boundary.decode()}, but found "
                f"{bytes_in[:len(boundary)].decode()}"
            )

        bytes_to_scan = bytes_in[
            len(boundary) :
        ]  # first boundary not needed anymore

        #  find the next boundary
        next_boundary_idx = bytes_to_scan.find(boundary)
        if next_boundary_idx == -1:
            # not found. There is no part here(yet)
            return b"", bytes_in
        else:
            part_bytes = bytes_to_scan[:next_boundary_idx]
            rest = bytes_to_scan[next_boundary_idx:]
            return part_bytes, rest


class MultipartContentError(DICOMTrolleyError):
    pass



"""Models parsing things into study/series/instance structure"""
from typing import Any, DefaultDict, Dict, List, Sequence

from pydicom.dataset import Dataset

from dicomtrolley.core import (
    DICOMObject,
    DICOMObjectLevels,
    DICOMObjectReference,
    Instance,
    Series,
    Study,
)
from dicomtrolley.exceptions import DICOMTrolleyError


def flatten(dicom_object) -> List[DICOMObject]:
    """All nodes and as a flat list, study, series"""
    nodes = [dicom_object]
    for child in dicom_object.children():
        nodes = nodes + flatten(child)
    return nodes


class TreeNode(DefaultDict[Any, "TreeNode"]):
    """Recursive defaultdict with a 'data' property. Helps parse to tree structures.

    Examples
    --------
    >>> root = TreeNode()
    >>> root['study1']['series1']['instance1'].data = 'some instance info'
    >>> 'study1' in root
    True
    >>> root['study1']['series2'].data = 'some series data'
    >>> list(root['study1'].keys)
    ['series1', 'series2']
    """

    def __init__(self, data=None, allow_overwrite=True):
        """

        Parameters
        ----------
        data:
            Optional data to associate with this node
        allow_overwrite: bool, optional
            If False, will raise exception when overwriting data attribute
        """
        super().__init__(lambda: TreeNode(allow_overwrite=allow_overwrite))
        self._data = data
        self.allow_overwrite = allow_overwrite

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if self.allow_overwrite or not self.data:
            self._data = value
        else:
            raise ValueError("Overwriting data is not allowed")


class DICOMParseTree:
    """Models study/series/instance as a tree. Allows arbitrary branch insertions"""

    def __init__(self, root=None, allow_overwrite=False):
        """

        Parameters
        ----------
        root: TreeNode, optional
            Root node. Will contain all. Defaults to empty
        allow_overwrite: bool, optional
            If False, will raise exception when overwriting any TreeNode.data,
            Defaults to False
        """
        if not root:
            self.root = TreeNode(allow_overwrite=allow_overwrite)

    @classmethod
    def init_from_objects(cls, objects: Sequence[DICOMObject]):
        """Create a tree from given objects. Useful for augmenting query results

        Notes
        -----
        Parent objects for instances and series will be created to maintain to
        ensure all nodes are connected to the tree root. For example, running this
        method with only a single instance as input will result in a regular
        root->study->series->instance tree. This is possible because all DICOMObjects
        maintain links to their parent objects

        Returns
        -------
        DICOMParseTree
        """

        tree = cls(allow_overwrite=True)
        for item in objects:
            tree.update(item)

        return tree

    def update(self, other):
        """Add all nodes in other to this tree. Overwrite existing

        Parameters
        ----------
        other: DICOMObject
            Dicom object to add to this tree
        """
        for item in flatten(other):
            self.insert_dicom_object(item)

    def insert(self, data, study_uid, series_uid=None, instance_uid=None):
        """Insert data at the given level in tree. Will complain about missing
        branches

        Raises
        ------
        DICOMTrolleyError
            If inserting fails for any reason
        """
        if instance_uid and not series_uid:
            DICOMTrolleyError(
                f"Instance was given ({instance_uid}) but series was not. I can "
                f"not insert this into a study/series/instance tree"
            )
        try:
            if series_uid:
                if instance_uid:
                    self.root[study_uid][series_uid][instance_uid].data = data
                else:
                    self.root[study_uid][series_uid].data = data
            else:
                self.root[study_uid].data = data
        except ValueError as e:
            raise DICOMTrolleyError(
                f"Error inserting dataset into "
                f"{study_uid}/{series_uid}/{instance_uid}: {e}"
            ) from e

    def insert_dataset(self, ds: Dataset):
        self.insert(
            data=ds,
            study_uid=ds.get("StudyInstanceUID"),
            series_uid=ds.get("SeriesInstanceUID"),
            instance_uid=ds.get("SOPInstanceUID"),
        )

    def insert_dicom_object(self, dicom_object: DICOMObject):
        """Insert dicomtrolley dicom object"""

        if isinstance(dicom_object, Study):
            self.insert(data=dicom_object.data, study_uid=dicom_object.uid)
        elif isinstance(dicom_object, Series):
            self.insert(
                data=dicom_object.data,
                study_uid=dicom_object.parent.uid,
                series_uid=dicom_object.uid,
            )
        elif isinstance(dicom_object, Instance):
            self.insert(
                data=dicom_object.data,
                study_uid=dicom_object.parent.parent.uid,
                series_uid=dicom_object.parent.uid,
                instance_uid=dicom_object.uid,
            )
        else:
            raise DICOMTrolleyError(
                f"Unknown DICOM object {dicom_object}. I do "
                f"not know where to insert this in the DICOM "
                f"tree"
            )

    @staticmethod
    def as_study(study_node_in) -> Study:
        """Tree node at the study level into a Study containing Series, Instances"""

        def value_or_dataset(val):
            if val:
                return val
            else:
                return Dataset()

        study_instance_uid, study_node = study_node_in

        study = Study(
            uid=study_instance_uid,
            data=value_or_dataset(study_node.data),
            series=tuple(),
        )
        all_series = []
        for series_instance_uid, series_node in study_node.items():
            series = Series(
                uid=series_instance_uid,
                data=value_or_dataset(series_node.data),
                parent=study,
                instances=tuple(),
            )
            all_series.append(series)
            all_instances = []
            for sop_instance_uid, instance_node in series_node.items():
                all_instances.append(
                    Instance(
                        uid=sop_instance_uid,
                        data=value_or_dataset(instance_node.data),
                        parent=series,
                    )
                )
            series.instances = tuple(all_instances)
        study.series = tuple(all_series)
        return study

    def as_studies(self) -> List[Study]:
        """Tree as DICOMObject instances: Study containing Series, Instance"""

        return [self.as_study(node) for node in self.root.items()]


class DICOMObjectNotFound(DICOMTrolleyError):
    pass


class DICOMObjectTree:
    """For querying, adding and modifying collections of DICOM objects

    Serves a different purpose from parsing.TreeNode. parsing.TreeNode is meant
    to build internally coherent(with parents, children) DICOMObjects from raw input.
    DICOMObjectTree allows manipulating these objects when they have already been
    created

    Notes
    -----
    If series or instances are added to the store, parent study or series
    will be created and added automatically.
    """

    def __init__(self, objects: Sequence[DICOMObject]):
        # make sure the tree is complete and all objects are based in studies
        studies = DICOMParseTree.init_from_objects(objects).as_studies()
        self._study_dict: Dict[str, Study] = {x.uid: x for x in studies}

    def __getitem__(self, series_uid) -> Study:
        return self._study_dict[series_uid]

    @property
    def studies(self):
        return list(self._study_dict.values())

    def add_study(self, study: Study):
        """Add given study to tree, overwriting exiting"""
        self._study_dict[study.uid] = study

    def retrieve(self, reference: DICOMObjectReference) -> DICOMObject:
        """Retrieve data for the given study, series or instance from tree

        Raises
        ------
        DICOMObjectNotFound
            If data for the given parameters does not exist in tree
        """

        try:
            if reference.level == DICOMObjectLevels.STUDY:
                return self[reference.study_uid]
            elif reference.level == DICOMObjectLevels.SERIES:
                return self[reference.study_uid].get(reference.series_uid)
            elif reference.level == DICOMObjectLevels.INSTANCE:
                return (
                    self[reference.study_uid]
                    .get(reference.series_uid)
                    .get(reference.instance_uid)
                )
            else:
                raise DICOMTrolleyError(
                    f"Unknown object level {reference.level}"
                )
        except KeyError as e:
            raise DICOMObjectNotFound(
                f"Study with uid {reference.study_uid} not found in tree"
            ) from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dicomtrolley']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'pydicom>=2.1.2,<3.0.0',
 'pynetdicom>=1.5.6,<2.0.0',
 'requests-futures>=1.0.0,<2.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'dicomtrolley',
    'version': '1.1.0',
    'description': 'Retrieve medical images via WADO, MINT, RAD69 and DICOM-QR',
    'long_description': '# dicomtrolley\n\n[![CI](https://github.com/sjoerdk/dicomtrolley/actions/workflows/build.yml/badge.svg?branch=master)](https://github.com/sjoerdk/dicomtrolley/actions/workflows/build.yml?query=branch%3Amaster)\n[![PyPI](https://img.shields.io/pypi/v/dicomtrolley)](https://pypi.org/project/dicomtrolley/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/dicomtrolley)](https://pypi.org/project/dicomtrolley/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n\nRetrieve medical images via WADO, MINT, RAD69 and DICOM-QR\n\n* Requires python 3.7, 3.8 or 3.9\n* Uses `pydicom` and `pynetdicom`. Images and query results are `pydicom.Dataset` instances\n* Multi-threaded downloading using `requests-futures`\n\n![A trolley](docs/resources/trolley.png)\n\n## Installation\n```\npip install dicomtrolley\n``` \n\n## Usage\n\n### Basic example\n\n```python\n# Create a logged-in http session\nsession = VitreaConnection(\n    "https://server/login").log_in(user,password,realm)\n                           \n# Use this session to create a trolley using MINT and WADO\ntrolley = Trolley(searcher=Mint(session, "https://server/mint"),\n                  wado=Wado(session, "https://server/wado"]))\n\n# find some studies (using MINT)\nstudies = trolley.find_studies(MintQuery(patientName=\'B*\'))  \n\n# download the fist one (using WADO)\ntrolley.download(studies[0], output_dir=\'/tmp/trolley\')\n```\n\n### Finding studies\n\n```python\nstudies = trolley.find_studies(MintQuery(patientName=\'B*\'))\n```\n\nQuery parameters can be found in [mint.Query](dicomtrolley/mint.py#L122). Valid include fields (which information gets sent back) can be found in [fields.py](dicomtrolley/fields.py):\n\n```python\nstudies = trolley.find_studies(\n    MintQuery(modalitiesInStudy=\'CT*\', \n              patientSex="F", \n              minStudyDate=datetime(year=2015, month=3, day=1),\n              maxStudyDate=datetime(year=2020, month=3, day=1),\n              includeFields=[\'PatientBirthDate\', \'SOPClassesInStudy\']))\n```\n\n### Finding series and instance details\nTo include series and instance level information as well, use the `queryLevel` parameter\n\n```python\nstudies = trolley.find_studies(  # find studies series and instances\n    MintQuery(studyInstanceID=\'B*\', \n              queryLevel=QueryLevels.INSTANCE)\n\na_series = studies.series[0]  # studies now contain series    \nan_instance = a_series.instances[0]  # and series contain instances\n```\n\n### Downloading data\nAny study, series or instance can be downloaded\n```python\nstudies = trolley.find_studies(MintQuery(patientName=\'B*\',\n                                         queryLevel=QueryLevels.INSTANCE))\n\npath = \'/tmp/trolley\'\ntrolley.download(studies, path)                             # all studies\ntrolley.download(studies[0]), path                          # a single study\ntrolley.download(studies[0].series[0], path)                # a single series\ntrolley.download(studies[0].series[0].instances[:3], path)  # first 3 instances\n```\nMore control over download: obtain `pydicom.Dataset` instances directly \n\n```python\nstudies = trolley.find_studies(              # find study including instances\n    Query(PatientID=\'1234\', \n          queryLevel=QueryLevels.INSTANCE)\n\nfor ds in trolley.get_dataset(studies):      # obtain Dataset for each instance\n    ds.save_as(f\'/tmp/{ds.SOPInstanceUID}.dcm\')\n```\n\nMulti-threaded downloading\n\n```python\ntrolley.download(studies, path, \n                 use_async=True,  # enable multi-threaded downloading \n                 max_workers=4)   # optionally set number of concurrent workers\n                                  # defaults to None which lets python decide\n```\n\nUsing WADO only, without search\n\n```python\nfrom dicomtrolley.wado import Wado\nfrom dicomtrolley.core import InstanceReference\n\ninstance = InstanceReference(series_instance_uid=\'1.2.1\', study_instance_uid=\'1.2.2\', sop_instance_uid=\'1.2.3\')\n\nwado = Wado(session, wado_url)\nfor ds in wado.datasets([instance]):\n  ds.save_as(f\'/tmp/{ds.SOPInstanceUID}.dcm\')\n```\n\n### DICOM-QR\n`Trolley` can use DICOM-QR instead of MINT as a search method. See [dicom_qr.DICOMQuery](dicomtrolley/dicom_qr.py#L30) for query details.\n```python\ndicom_qr = DICOMQR(host,port,aet,aec)\ntrolley = Trolley(searcher=dicom_qr, downloader=wado)\n\n# Finding is similar to MINT, but a DICOMQuery is used instead\ntrolley.find_studies(  \n    query=DICOMQuery(PatientName="BAL*",   \n                     minStudyDate=datetime(year=2015, month=3, day=1),\n                     maxStudyDate=datetime(year=2015, month=4, day=1),\n                     includeFields=["PatientBirthDate", "SOPClassesInStudy"],\n                     QueryRetrieveLevel=QueryRetrieveLevels.STUDY)) \n```\n\n### RAD69\nThe [RAD69](https://gazelle.ihe.net/content/rad-69-retrieve-imaging-document-set) protocol is an alternative to wado for downloading DICOM images.\n```python\ndicom_qr = DICOMQR(host,port,aet,aec)\ntrolley = Trolley(searcher=dicom_qr, \n                  downloader=Rad69(session=session,\n                                   url="https://server/rad69"))\n\nstudies = trolley.find_studies(PatientName="AB*",)\ntrolley.download(studies[0], path)  # rad69 download works exactly like wado \ntrolley.download(studies[1], path,  \n                 use_async=True)    # multi-threaded download is supported\n\n```\n## Examples\n* [Search for studies in MINT](examples/search_for_studies_mint.py) \n* [Search for studies in DICOM-QR](examples/search_for_studies_dicom_qr.py)\n* [Find and download studies](examples/go_shopping.py)\n* [Using WADO only](examples/use_wado_only.py)\n* [Download studies with rad69](examples/go_shopping_rad69.py)\n\n## Alternatives\n* [dicomweb-client](https://github.com/MGHComputationalPathology/dicomweb-client) - Active library supporting QIDO-RS, WADO-RS and STOW-RS. \n* [pynetdicom](https://github.com/pydicom/pynetdicom) - dicomtrolley\'s DICOM-QR support is based on pynetdicom. Pynetdicom supports a broad range of DICOM networking interactions and can be used as a stand alone application.\n\n## Caveats\nDicomtrolley has been developed for and tested on a Vitrea Connection 8.2.0.1 system. This claims to\nbe consistent with WADO and MINT 1.2 interfaces, but does not implement all parts of these standards. \n\nCertain query parameter values and restraints might be specific to Vitrea Connection 8.2.0.1. For example,\nthe exact list of DICOM elements that can be returned from a query might be different for different servers.\n\n\n# Contributing\nYou can contribute in different ways\n\n## Report bugs\nReport bugs at https://github.com/sjoerdk/clockify_api_client/issues.\n\n## Contribute code\n### Get the code\nFork this repo, create a feature branch\n\n### Set up environment\ndicomtrolley uses [poetry](https://python-poetry.org/docs/) for dependency and package management \n* Install poetry (see [poetry docs](https://python-poetry.org/docs/#installation))\n* Create a virtual env. Go to the folder where cloned dicomtrolley and use \n  ```  \n  poetry install \n  ``` \n* Install [pre-commit](https://pre-commit.com) hooks.\n  ```\n  pre-commit install\n  ```\n  \n### Add your code \nMake your code contributions. Make sure document and add tests for new features.\nTo automatically publish to pypi, increment the version number. See below. \n\n### Lint your code\n* Run all tests\n* Run [pre-commit](https://pre-commit.com):\n  ```\n  pre-commit run\n  ```\n### Publish\nCreate a pull request\n\n### Incrementing the version number\nA merged pull request will only be published to pypi if it has a new version number. \nTo bump dicomtrolley\'s version, do the following.\n* dicomtrolley uses [semantic versioning](https://semver.org/) Check whether your addition is a PATCH, MINOR or MAJOR version.\n* Manually increment the version number:\n  * `pyproject.toml` -> `version = "0.1.2"`\n  \n* Add a brief description of your updates new version to `HISTORY.md`\n',
    'author': 'sjoerdk',
    'author_email': 'sjoerd.kerkstra@radboudumc.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sjoerdk/dicomtrolley',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""XML templates to use with jinja2. Separate file so I can ignore flake8 line
too long linting errors.
"""

# SOAP structure to POST to a rad69 endpoint to retrieve a single slice. A beauty.
RAD69_SOAP_REQUEST_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
                                    <s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://www.w3.org/2005/08/addressing">
                                        <s:Header>
                                            <a:Action s:mustUnderstand="1">urn:ihe:rad:2009:RetrieveImagingDocumentSet </a:Action>
                                            <a:MessageID>urn:uuid:{{ uuid }}</a:MessageID>
                                            <a:ReplyTo s:mustUnderstand="1">
                                                <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
                                            </a:ReplyTo>
                                            <a:To >http://localhost:2647/XdsService/IHEXDSIDocSource.svc</a:To>
                                        </s:Header>
                                        <s:Body>response
                                            {% for study in studies %}
                                                <iherad:RetrieveImagingDocumentSetRequest xmlns:iherad="urn:ihe:rad:xdsi-b:2009" xmlns:ihe="urn:ihe:iti:xds-b:2007">
                                                    <iherad:StudyRequest studyInstanceUID="{{ study.uid }}">
                                                        {% for series in study.series %}
                                                        <iherad:SeriesRequest seriesInstanceUID="{{ series.uid }}">
                                                            {% for instance in series.instances  %}
                                                            <ihe:DocumentRequest>
                                                                <ihe:RepositoryUniqueId>1.3.6.1.4.1000</ihe:RepositoryUniqueId>
                                                                <ihe:DocumentUniqueId>{{ instance.uid }}</ihe:DocumentUniqueId>
                                                            </ihe:DocumentRequest>
                                                            {% endfor %}
                                                        </iherad:SeriesRequest>
                                                        {% endfor %}
                                                    </iherad:StudyRequest>
                                                    <iherad:TransferSyntaxUIDList> {% for transfer_syntax_id in transfer_syntax_list %}
                                                        <iherad:TransferSyntaxUID>{{ transfer_syntax_id }}</iherad:TransferSyntaxUID> {% endfor %}
                                                    </iherad:TransferSyntaxUIDList>
                                                </iherad:RetrieveImagingDocumentSetRequest>
                                            {% endfor %}
                                            </s:Body>
                                    </s:Envelope>
                                """


A_RAD69_RESPONSE_SOAP_HEADER = """<?xml version="1.0" encoding="utf-8" ?>
                                    <env:Envelope xmlns:env="http://www.w3.org/2003/05/soap-envelope">
                                        <env:Header>
                                            <a:Action xmlns:a="http://www.w3.org/2005/08/addressing env:mustUnderstand="true">urn:ihe:iti:2007:RetrieveDocumentSetResponse</a:Action>
                                            <a:RelatesTo xmlns:a="http://www.w3.org/2005/08/addressing">urn:uuid:1.2.3</a:RelatesTo>
                                        </env:Header>
                                        <env:Body>
                                            <RetrieveDocumentSetResponse xmlns="urn:ihe:iti:xds-b:2007 xmlns:rs="urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0">
                                                <rs:RegistryResponse status="urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success"/>
                                                <DocumentResponse>
                                                    <RepositoryUniqueId>1.3.6.1.4.1000</RepositoryUniqueId>
                                                    <DocumentUniqueId>1.3.12.2.1107.5.2.19.45030.2015100517150896126425886</DocumentUniqueId>
                                                    <HomeCommunityId/>
                                                    <mimeType>application/dicom</mimeType>
                                                    <Document><xop:Include xmlns:xop="http://www.w3.org/2004/08/xop/include href="cid:e6f6b279-b259-436e-8d9e-73df1d1157d1"/></Document>
                                                </DocumentResponse>
                                            </RetrieveDocumentSetResponse>
                                        </env:Body>
                                    </env:Envelope>"""

A_RAD69_RESPONSE_SOAP_HEADER_TEMPLATE = """<?xml version="1.0" encoding="utf-8" ?>
                                    <env:Envelope 
                                    xmlns:env="http://www.w3.org/2003/05/soap-envelope">
                                        <env:Header>
                                            <a:Action 
                                            xmlns:a="http://www.w3.org/2005/08/addressing env:mustUnderstand="true">urn:ihe:iti:2007:RetrieveDocumentSetResponse</a:Action>
                                            <a:RelatesTo 
                                            xmlns:a="http://www.w3.org/2005/08/addressing">urn:uuid:1.2.3</a:RelatesTo>
                                        </env:Header>
                                        <env:Body>
                                            <RetrieveDocumentSetResponse 
                                            xmlns="urn:ihe:iti:xds-b:2007 
                                            xmlns:rs="urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0">
                                                <rs:RegistryResponse 
                                                status="urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success"/>
                                                {% for did in dids %}
                                                <DocumentResponse>
                                                    <RepositoryUniqueId>1.3.6.1.4.1000</RepositoryUniqueId>
                                                    <DocumentUniqueId>{{ did }}</DocumentUniqueId>
                                                    <HomeCommunityId/>
                                                    <mimeType>application/dicom</mimeType>
                                                    <Document><xop:Include 
                                                    xmlns:xop="http://www.w3.org/2004/08/xop/include href="cid:e6f6b279-b259-436e-8d9e-73df1d1157d1"/></Document>
                                                </DocumentResponse>
                                                {% endfor %}
                                            </RetrieveDocumentSetResponse>
                                        </env:Body>
                                    </env:Envelope>"""

# the html response contents of a successful rad69 response
RAD69_SOAP_RESPONSE_TEMPLATE = """------=_Part_1_1788103738.1647962762356
                                  Content-Type: application/xop+xml; charset=utf-8; type="application/soap+xml"
                                  Content-Id: <ff8941e7-4b9f-44a5-b71c-338a33547f03>
                                  Content-Transfer-Encoding: binary

                                 <?xml version="1.0" encoding="utf-8" ?>
                                 <env:Envelope xmlns:env=http://www.w3.org/2003/05/soap-envelope">
                                    <env:Header>
                                        <a:Action xmlns:a="http://www.w3.org/2005/08/addressing" env:mustUnderstand="true">urn:ihe:iti:2007:RetrieveDocumentSetResponse</a:Action>
                                        <a:RelatesTo xmlns:a="http://www.w3.org/2005/08/addressing">urn:uuid:1.2.3</a:RelatesTo>
                                    </env:Header>
                                    <env:Body>
                                        <RetrieveDocumentSetResponse xmlns="urn:ihe:iti:xds-b:2007 xmlns:rs="urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0">
                                            <rs:RegistryResponse status="urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success"/>
                                                <DocumentResponse>
                                                    <RepositoryUniqueId>1.3.6.1.4.1000</RepositoryUniqueId>
                                                    <DocumentUniqueId>1.2.3.4.5.6.7</DocumentUniqueId>
                                                    <HomeCommunityId/>
                                                    <mimeType>application/dicom</mimeType>
                                                    <Document><xop:Include xmlns:xop="http://www.w3.org/2004/08/xop/include" href="cid:e6f6b279-b259-436e-8d9e-73df1d1157d1"/>
                                                    </Document>
                                                </DocumentResponse>
                                        </RetrieveDocumentSetResponse>
                                    </env:Body>
                                </env:Envelope>
                                ------=_Part_1_1788103738.1647962762356
                                Content-Type: application/dicom
                                Content-ID: e6f6b279-b259-436e-8d9e-73df1d1157d1

                                {{ dicom_bytestream }}

                                ------=_Part_1_1788103738.1647962762356--"""

# what a rad69 server sends back when requesting a non-existant slice
RAD69_SOAP_RESPONSE_NOT_FOUND = """<?xml version="1.0" encoding="utf-8" ?>
                                        <env:Envelope xmlns:env="http://www.w3.org/2003/05/soap-envelope">
                                            <env:Header>
                                                <a:Action xmlns:a="http://www.w3.org/2005/08/addressing" env:mustUnderstand="true">urn:ihe:iti:2007:RetrieveDocumentSetResponse</a:Action>
                                                <a:RelatesTo xmlns:a="http://www.w3.org/2005/08/addressing">urn:uuid:67bce40e-52ce-450e-8ef5-bd093d4b2dfd</a:RelatesTo>
                                            </env:Header>
                                            <env:Body>
                                                <RetrieveDocumentSetResponse xmlns="urn:ihe:iti:xds-b:2007" xmlns:rs="urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0">
                                                    <rs:RegistryResponse status="urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Failure">
                                                        <rs:RegistryErrorList>
                                                            <rs:RegistryError codeContext="Unable to locate SOP from MINT." errorCode="XDSMissingDocument" location="1.2.40.0.13.1.16930207878723102225780279777977199534.1" severity="urn:oasis:names:tc:ebxml-regrep:ErrorSeverityType:Error"/>
                                                            <rs:RegistryError codeContext="Unable to locate SOP from MINT." errorCode="XDSMissingDocument" location="1.2.40.0.13.1.58933371510581919046064437805732852473.1" severity="urn:oasis:names:tc:ebxml-regrep:ErrorSeverityType:Error"/>
                                                        </rs:RegistryErrorList>
                                                    </rs:RegistryResponse>
                                                </RetrieveDocumentSetResponse>
                                            </env:Body>
                                        </env:Envelope>
"""

# Xpath query to retrieve errors from a rad69 soap response. Nice and readable.
RAD69_SOAP_RESPONSE_ERROR_XPATH = "./{http://www.w3.org/2003/05/soap-envelope}Body/{urn:ihe:iti:xds-b:2007}RetrieveDocumentSetResponse/{urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0}RegistryResponse/{urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0}RegistryErrorList/*"
